from sqlalchemy import String, Text
from sqlalchemy.orm import Mapped, mapped_column

from .base_model import InternalBaseModel


class BaseHistoryModel(InternalBaseModel):
    __abstract__ = True

    operator_id: Mapped[str] = mapped_column(String(100))
    operator_name: Mapped[str] = mapped_column(String(100))
    operator_type: Mapped[str] = mapped_column(String(100))
    operate: Mapped[str] = mapped_column(Text(2000))



import contextlib
from typing import AsyncIterator

from sqlalchemy.engine import URL
from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine, AsyncConnection
from sqlalchemy.orm import sessionmaker

from .exception.internal_exception import DatabaseInitializeFailureException, DatabaseConnectFailureException


class MariaDB:

    def __init__(self, user_name: str, password: str, host: str, port: int, db_name: str):
        self._connection_url = URL.create("mysql+aiomysql",
                                          username=user_name,
                                          password=password,
                                          host=host,
                                          port=port,
                                          database=db_name)
        self._engine = create_async_engine(self._connection_url, echo=True, future=True, pool_pre_ping=True)
        self._session_local = sessionmaker(class_=AsyncSession, autocommit=False, autoflush=False, bind=self._engine,
                                           expire_on_commit=False)

    async def close(self):
        if self._engine:
            await self._engine.dispose()

            self._engine = None
            self._session_local = None

    @contextlib.asynccontextmanager
    async def connect(self) -> AsyncIterator[AsyncConnection]:
        if self._engine is None:
            raise DatabaseInitializeFailureException()

        async with self._engine.begin() as connection:
            try:
                yield connection
            except Exception:
                await connection.rollback()
                raise DatabaseConnectFailureException()

    @contextlib.asynccontextmanager
    async def session(self) -> AsyncIterator[AsyncSession]:
        if self._session_local is None:
            raise DatabaseConnectFailureException()

        session = self._session_local()
        try:
            yield session
            await session.commit()
        except Exception:
            await session.rollback()
            raise
        finally:
            await session.close()



import logging
import logging.handlers
import os
import traceback
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager
from functools import lru_cache
from asgi_correlation_id import CorrelationIdMiddleware, CorrelationIdFilter

import dotenv
import watchtower
from fastapi import FastAPI, status, Request
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware

from sqlalchemy.engine import URL
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.jobstores.sqlalchemy import SQLAlchemyJobStore

from . import database
from .const import LOG_FMT, LOG_FMT_NO_DT, LOG_DT_FMT, DEFAULT_LOGGER_NAME, APS_EVENT_CODE, \
    CORRELATION_ID_HEADER_KEY_NAME
from .exception.base_exception import InternalBaseException
from .exception.internal_exception import BadGatewayException, GatewayTimeoutException
from .ext.amazon import aws
from .http.requests import send_webhook_message
from .http.responses import async_response
from .utils import update_dict_with_cast


class BaseFactory(metaclass=ABCMeta):
    DEFAULT_APP_NAME = ""
    API_VERSION = "v0.0.0"

    @abstractmethod
    def init_modules(self, app):
        """
        Each factory should define what modules it wants.
        """

    @abstractmethod
    def init_modules_job(self, app):
        """
        Each factory should define what modules job it wants.
        """

    @abstractmethod
    @lru_cache()
    def get_app_config(self):
        """
        Each factory should define what config it wants.
        """

    def create_app(self, title=None) -> FastAPI:
        lifespan = None

        @asynccontextmanager
        async def lifespan(app: FastAPI):

            await self.__init_apscheduler(app)

            yield
            if app.state.db._engine is not None:
                await app.state.db.close()

            app.state.scheduler.shutdown()
            app.state.logger.warn("Apscheduler shutdown")

        if title is None:
            title = self.DEFAULT_APP_NAME

        if self.get_app_config().DEBUG:
            app = FastAPI(openapi_url=self.get_app_config().OPEN_API_URL, title=title,
                          debug=self.get_app_config().DEBUG,
                          version=self.API_VERSION, lifespan=lifespan)
        else:
            app = FastAPI(openapi_url=self.get_app_config().OPEN_API_URL, title=title,
                          debug=self.get_app_config().DEBUG,
                          version=self.API_VERSION, lifespan=lifespan, docs_url=None, redoc_url=None)

        origins = ["*"]

        app.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
            expose_headers=[CORRELATION_ID_HEADER_KEY_NAME]
        )

        app.add_middleware(
            CorrelationIdMiddleware,
            header_name=CORRELATION_ID_HEADER_KEY_NAME,
            update_request_header=True
        )

        self.__load_local_config()
        app.state.config = self.get_app_config()
        self.__setup_main_logger(app, level=logging.DEBUG)
        app.state.aws_session = aws.init_app(app)
        self.__setup_cloud_log(app)
        self.__load_cloud_config(app)

        db = database.MariaDB(self.get_app_config().DATABASE_USERNAME, self.get_app_config().DATABASE_PASSWORD,
                              self.get_app_config().DATABASE_HOST, self.get_app_config().DATABASE_PORT,
                              self.get_app_config().DATABASE_NAME)

        app.state.db = db
        app.state.config = self.get_app_config()
        self.__init_modules(app)
        self.__init_builtin_api(app)

        @app.exception_handler(InternalBaseException)
        async def http_exception_handler(request: Request, exc: InternalBaseException):
            detail = exc.detail

            if isinstance(exc, BadGatewayException):
                message = f"【{self.DEFAULT_APP_NAME}】Bad gateway, request:{request.__dict__}, exc:{exc}"
                await send_webhook_message(app, message)
            elif isinstance(exc, GatewayTimeoutException):
                message = f"【{self.DEFAULT_APP_NAME}】Gateway timeout, request:{request.__dict__}, exc:{exc}"
                await send_webhook_message(app, message)

            return await async_response(data=detail.get("data"), code=detail.get("code"), message=detail.get("message"),
                                        status_code=exc.status_code)

        @app.exception_handler(RequestValidationError)
        async def validation_exception_handler(request: Request, exc: RequestValidationError):
            data = {"detail": exc.errors()}
            if exc.body:
                data["body"] = exc.body
            return await async_response(data=data,
                                        code="error_unprocessable_entity", message="參數或格式錯誤",
                                        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY)

        @app.exception_handler(Exception)
        async def http_exception_handler(request: Request, exc: Exception):
            app.state.logger.warn(f"Exception, request:{request.__dict__}, exc:{exc}")
            app.state.logger.warn(traceback.format_exc())
            message = f"【{self.DEFAULT_APP_NAME}】Unprocessed Exception, request:{request.__dict__}, exc:{exc}"
            await send_webhook_message(app, message)

            return await async_response(code="error_internal_server", message="系統錯誤",
                                        status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)

        return app

    def __load_local_config(self):
        dotenv.load_dotenv(override=True)
        update_dict_with_cast(self.get_app_config(), os.environ)

    def __load_cloud_config(self, app):
        if not app.state.aws_session or not self.get_app_config().AWS_PARAMETER_PATH_PREFIX:
            app.state.logger.warn("No AWS session or Parameter Storage configuration, ignore cloud config")
            return

        cloud_conf = {}

        params = {
            "Path": self.get_app_config().AWS_PARAMETER_PATH_PREFIX,
            "Recursive": True,
            "WithDecryption": True
        }

        # AWS only give us 10 parameters per api call
        ssm_client = app.state.aws_session.client("ssm")
        while True:
            result = ssm_client.get_parameters_by_path(**params)
            cloud_conf.update({para["Name"].split("/")[-1]: para["Value"] for para in result["Parameters"]})
            if not result.get("NextToken"):
                break
            params.update({"NextToken": result["NextToken"]})

        update_dict_with_cast(self.get_app_config(), cloud_conf)

    def __init_modules(self, app):
        self.init_modules(app)

    def __setup_main_logger(self, app, logger_name=DEFAULT_LOGGER_NAME, level=logging.INFO):
        logger = self.__setup_logger(app, logger_name, level)
        app.state.logger = logger

    #
    def __setup_logger(self, app, logger_name, level=logging.INFO):
        logger = logging.getLogger(logger_name)
        logger.setLevel(level)

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(fmt=LOG_FMT))
        stream_handler.addFilter(CorrelationIdFilter())
        logger.addHandler(stream_handler)

        return logger

    def __setup_cloud_log(self, app):
        if app.state.aws_session and self.get_app_config().AWS_LOGGROUP_NAME:
            logs_client = app.state.aws_session.client("logs")
            watchtower_handler = watchtower.CloudWatchLogHandler(
                log_group_name=self.get_app_config().AWS_LOGGROUP_NAME,
                boto3_client=logs_client, create_log_group=False)
            watchtower_handler.setFormatter(logging.Formatter(fmt=LOG_FMT_NO_DT, datefmt=LOG_DT_FMT))
            watchtower_handler.addFilter(CorrelationIdFilter())
            app.state.logger.addHandler(watchtower_handler)

            # 获取和设置 SQLAlchemy 的日志器
            sqlalchemy_logger = logging.getLogger('sqlalchemy.engine')
            sqlalchemy_logger.setLevel(logging.DEBUG)  # 根据需要设置日志级别
            sqlalchemy_logger.addHandler(watchtower_handler)

    def __init_builtin_api(self, app):

        @app.get(f'/health', tags=["System"])
        async def health():
            return await async_response()

        @app.get(f'/hello', tags=["System"])
        async def hello():
            return await async_response(data={"API version": app.version})

    async def __init_apscheduler(self, app):
        sql_url = URL.create("mysql+pymysql",
                             username=self.get_app_config().DATABASE_USERNAME,
                             password=self.get_app_config().DATABASE_PASSWORD,
                             host=self.get_app_config().DATABASE_HOST,
                             port=self.get_app_config().DATABASE_PORT,
                             database=self.get_app_config().DATABASE_NAME)
        sql_url = sql_url.render_as_string(False)
        scheduler = AsyncIOScheduler(jobstores={'default': SQLAlchemyJobStore(url=sql_url)})
        app.state.scheduler = scheduler
        app.state.logger.info("Apscheduler initialized")
        self.init_modules_job(app)
        app.state.logger.info("Modules job initialized")

        from apscheduler.events import EVENT_JOB_ADDED, EVENT_JOB_ERROR, EVENT_JOB_EXECUTED, EVENT_JOB_MISSED, \
            EVENT_JOB_REMOVED, EVENT_JOB_SUBMITTED

        def job_event(event):
            app.state.logger.info(f'Job event occur, job: {event.job_id}, event: {APS_EVENT_CODE.get(event.code)}')

        app.state.scheduler.add_listener(
            job_event,
            EVENT_JOB_ADDED | EVENT_JOB_ERROR | EVENT_JOB_EXECUTED | EVENT_JOB_MISSED | EVENT_JOB_REMOVED | EVENT_JOB_SUBMITTED
        )

        app.state.scheduler.start()



from dataclasses import dataclass, field, asdict, fields
from typing import Optional


@dataclass
class InternalBaseInterface:
    def __post_init__(self):
        # 初始化 __explicitly_set_fields__ 集合
        self.__explicitly_set_fields__ = set()

        for f in fields(self):
            value = getattr(self, f.name)
            if f.name in self.__explicitly_set_fields__ or value is not f.default:
                self.__explicitly_set_fields__.add(f.name)

    def __init__(self, **kwargs):
        # 初始化 __explicitly_set_fields__ 集合
        self.__explicitly_set_fields__ = set()

        for f in fields(self):
            if f.name in kwargs:
                value = kwargs[f.name]
                setattr(self, f.name, value)
                self.__explicitly_set_fields__.add(f.name)

    def to_dict(self, exclude_unset: bool = False, exclude_defaults: bool = False, exclude_none: bool = False):
        # 將 dataclass 轉換為字典
        data = asdict(self)

        if exclude_unset:
            # 過濾掉那些未被明確設置的欄位
            data = {k: v for k, v in data.items() if k in self.__explicitly_set_fields__ or v is not None}

        if exclude_defaults:
            # 過濾掉那些設置為預設值的欄位
            data = {k: v for k, v in data.items() if v != self.__dataclass_fields__[k].default}

        if exclude_none:
            # 過濾掉那些值為 None 的欄位
            data = {k: v for k, v in data.items() if v is not None}

        return data



from pydantic_settings import BaseSettings


class BaseConfig(BaseSettings):
    DEBUG: bool = False
    RUN_PORT: int = 5000
    TIME_ZONE: str = "Asia/Taipei"

    OPEN_API_URL: str = "/openapi.json"

    # Request
    REQUEST_VERIFY_SSL: bool = True
    REQUEST_PROXY: str = ''
    REQUEST_RETRY_COUNT: int = 0
    REQUEST_RETRY_DELAY_INITIAL_SECONDS: float = 1.0
    REQUEST_RETRY_DELAY_FACTOR: float = 1.5
    REQUEST_RETRY_DELAY_RANDOM_JITTER_MIN: float = 0.0
    REQUEST_RETRY_DELAY_RANDOM_JITTER_MAX: float = 0.5
    REQUEST_CONN_POOL_TIMEOUT: float = 5.0
    REQUEST_CONN_TIMEOUT: float = 5.0
    REQUEST_WRITE_TIMEOUT: float = 5.0
    RESPONSE_READ_TIMEOUT: float = 10.0

    # AWS
    AWS_ACCESS_KEY_ID: str = ""
    AWS_SECRET_KEY: str = ""
    AWS_REGION: str = ""
    AWS_PARAMETER_PATH_PREFIX: str = ""
    AWS_LOGGROUP_NAME: str = ""

    # MariaDB
    DATABASE_HOST: str = ''
    DATABASE_USERNAME: str = ''
    DATABASE_PASSWORD: str = ''
    DATABASE_PORT: int = 3306
    DATABASE_NAME: str = ""

    # Exception Notify
    WEBHOOK_BASE_URL: str = ""
    WEBHOOK_RETRY_COUNT: int = 5
    WEBHOOK_RETRY_DELAY_INITIAL_SECONDS: float = 1.0
    WEBHOOK_RETRY_DELAY_FACTOR: float = 2.0
    WEBHOOK_RETRY_DELAY_RANDOM_JITTER_MIN: float = 0.0
    WEBHOOK_RETRY_DELAY_RANDOM_JITTER_MAX: float = 0.5

    class Config:
        case_sensitive = False



from typing import Optional, Union

from pydantic import BaseModel


class PaginationResponse(BaseModel):
    page_data: Optional[list] = None
    page_no: int
    page_size: int
    total_num: int


class BaseResponse(BaseModel):
    code: int
    message: Optional[str] = None
    data: Optional[Union[dict, list]] = None



# -*- coding: UTF-8 -*-
import boto3

from .const import AWS_CONF_KEY


def init_app(app):
    if not all([getattr(app.state.config, key) for key in AWS_CONF_KEY]):
        # pylint: disable=no-member
        app.state.logger.warn("Lack AWS credential keys, ignore connect to AWS")
        return None

    aws_session = boto3.session.Session(
        aws_access_key_id=app.state.config.AWS_ACCESS_KEY_ID,
        aws_secret_access_key=app.state.config.AWS_SECRET_KEY,
        region_name=app.state.config.AWS_REGION
    )

    # This should be logging when create Logging handlers,
    # But we have too many CloudWatchLogHandler, only print once here.
    if not getattr(app.state.config, "AWS_LOGGROUP_NAME"):
        app.state.logger.info("Lack AWS configuration keys, ignore AWS CloudWatch log handlers")

    return aws_session



import json
import httpx

from fastapi import status
from fastapi.responses import JSONResponse
from fastapi.encoders import jsonable_encoder

from ..model.base_model import InternalBaseModel


async def async_response(data=None, message=None, code=None, page_no=None, total_num=None, page_size=None,
                         status_code=status.HTTP_200_OK):
    def _serialize(data):
        if isinstance(data, InternalBaseModel):
            data = json.loads(json.dumps(jsonable_encoder(data.model_dump()), ensure_ascii=False))
        return data

    if isinstance(data, httpx.Response):
        return JSONResponse(status_code=data.status_code, content=data.json())

    ret = {}
    if isinstance(data, list):
        data = [_serialize(d) for d in data]
    else:
        data = _serialize(data)

    data = jsonable_encoder(data)

    ret['code'] = code or "ok"

    ret['message'] = message or "success"

    if page_no and total_num and page_size:
        ret['data'] = {
            'page_no': page_no,
            'page_size': page_size,
            'total_num': total_num,
            'page_data': data
        }
    else:
        ret['data'] = data

    return JSONResponse(status_code=status_code, content=ret)



from fastapi import HTTPException


class InternalBaseException(HTTPException):
    def __init__(self, status_code: int = None, code: str = "ok", message: str = "success", **kwargs):
        detail = {
            "code": code,
            "message": message,
            "data": kwargs,
        }
        super().__init__(status_code=status_code, detail=detail)



import copy
import json
import uuid
import arrow
from datetime import datetime
from typing import Union, Dict, Type, Tuple, Set, Mapping, Any, TYPE_CHECKING

from sqlalchemy.ext.asyncio.session import AsyncSession

from sqlalchemy import Column, DateTime, String
from sqlalchemy.ext.declarative import as_declarative, declared_attr
from sqlalchemy.ext.asyncio import AsyncAttrs
from pydantic import BaseModel

from .operate import Operate
from ..const import STR_DELIMIT
from ..exception.internal_exception import NoChangeException


@as_declarative()
class Base(AsyncAttrs):
    __abstract__ = True
    __name__: str

    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()


class InternalBaseModel(Base):
    __abstract__ = True

    id = Column(String(50), primary_key=True, nullable=False, default=uuid.uuid4)
    create_time = Column(DateTime, default=datetime.now)
    update_time = Column(DateTime, default=datetime.now, onupdate=datetime.now)

    def model_dump(self, mode: str = "json", exclude: set = None) -> Dict[str, Any]:
        """將 SQLAlchemy 模型轉換為字典"""
        exclude = exclude or set()
        result = {}

        for column in self.__table__.columns:
            if column.name not in exclude:
                value = getattr(self, column.name)
                if mode == "json" and isinstance(value, datetime):
                    result[column.name] = value.isoformat()
                else:
                    result[column.name] = value

        return result

    def model_copy(self):
        return copy.deepcopy(self.model_dump())

    async def update_wrap(self, db: AsyncSession, schema: Union[Dict, Type[BaseModel]], current_operator,
                          history_model=None) -> 'InternalBaseModel':
        from .base_history_model import BaseHistoryModel

        if history_model and not issubclass(type(history_model), BaseHistoryModel):
            raise TypeError("history_model must be a subclass of BaseHistoryModel")

        if not issubclass(type(schema), dict) and not issubclass(type(schema), BaseModel):
            raise TypeError("Schema must be a subclass of BaseModel or dict")

        original_model = self.model_copy()
        delta_dict = schema
        if issubclass(type(schema), BaseModel):
            delta_dict = schema.model_dump(exclude_unset=True, mode="json")

        for key, value in delta_dict.items():
            if hasattr(self, key):
                if isinstance(value, list):
                    value = STR_DELIMIT.join(value) if value else None

                setattr(self, key, value)

        operate = await Operate.generate_operate(original_model, self.model_dump())
        if not operate.add and not operate.remove and not operate.change:
            await db.flush()
            await db.refresh(self)
            return self

        await db.flush()
        await db.refresh(self)

        if history_model:
            if hasattr(history_model, f"{self.__tablename__}_id"):
                setattr(history_model, f"{self.__tablename__}_id", self.id)

            history_model.operator_id = current_operator.id
            history_model.operator_name = current_operator.name
            history_model.operator_type = current_operator.type
            history_model.operate = json.dumps(operate.model_dump(), ensure_ascii=False, indent=4)

            db.add(history_model)
            await db.flush()
            await db.refresh(history_model)

        return self

    async def create_wrap(self, db: AsyncSession, schema: Union[Dict, Type[BaseModel]], current_operator,
                          history_model=None) -> 'InternalBaseModel':
        from .base_history_model import BaseHistoryModel

        if history_model and not issubclass(type(history_model), BaseHistoryModel):
            raise TypeError("history_model must be a subclass of BaseHistoryModel")

        if not issubclass(type(schema), dict) and not issubclass(type(schema), BaseModel):
            raise TypeError("Schema must be a subclass of BaseModel or dict")

        delta_dict = schema
        if issubclass(type(schema), BaseModel):
            delta_dict = schema.model_dump(exclude_unset=True, mode="json")

        for key, value in delta_dict.items():
            if hasattr(self, key):
                if isinstance(value, list):
                    value = STR_DELIMIT.join(value) if value else None

                setattr(self, key, value)

        operate = await Operate.generate_operate(compare=self.model_dump())

        db.add(self)
        await db.flush()
        await db.refresh(self)

        if history_model:
            if hasattr(history_model, f"{self.__tablename__}_id"):
                setattr(history_model, f"{self.__tablename__}_id", self.id)

            history_model.operator_id = current_operator.id
            history_model.operator_name = current_operator.name
            history_model.operator_type = current_operator.type
            history_model.operate = json.dumps(operate.model_dump(), ensure_ascii=False, indent=4)

            db.add(history_model)
            await db.flush()
            await db.refresh(history_model)

        return self



import asyncio
import random

import httpx

from fastapi import FastAPI
from fastapi.encoders import jsonable_encoder
from asgi_correlation_id import correlation_id

from ..const import CORRELATION_ID_HEADER_KEY_NAME
from ..exception.internal_exception import GatewayTimeoutException, BadGatewayException


async def invoke_request(timeout: httpx.Timeout, method: str, url: str, app: FastAPI, **kwargs):
    try:
        async with httpx.AsyncClient(timeout=timeout, verify=app.state.config.REQUEST_VERIFY_SSL) as client:
            if "json" in kwargs:
                kwargs["json"] = jsonable_encoder(kwargs["json"])

            app.state.logger.info(f"【{app.title}】 \nURL: {method} {url} \nkwargs: {kwargs} \ninvoke_request() request")
            response = await client.request(method, url, **kwargs)
            app.state.logger.info(
                f"【{app.title}】 \nURL: {method} {url} \nkwargs: {kwargs} \ninvoke_request() response \n\nstatus_code: {response.status_code} \ntext: {response.text}"
            )
            return response
    except httpx.TimeoutException as exc:
        message = f"【{app.title}】 \nURL: {method} {url} \nkwargs: {kwargs} \ninvoke_request(), TimeoutException, exc: {exc}"
        app.state.logger.warn(message)
        raise GatewayTimeoutException(str(exc)) from exc
    except Exception as exc:
        message = f"【{app.title}】 \nURL: {method} {url} \nkwargs: {kwargs} \ninvoke_request(), Exception, exc: {exc}"
        app.state.logger.warn(message)
        raise BadGatewayException(str(exc))


async def async_request(app: FastAPI, method, url, current_user: dict = None,
                        request_conn_pool_timeout: float = -1.0, request_conn_timeout: float = -1.0,
                        request_write_timeout: float = -1.0, response_read_timeout: float = -1.0,
                        request_retry_count: int = -1, request_retry_delay_initial_seconds: float = -1.0,
                        **kwargs):
    if request_conn_pool_timeout < 0:
        request_conn_pool_timeout = app.state.config.REQUEST_CONN_POOL_TIMEOUT
    if request_conn_timeout < 0:
        request_conn_timeout = app.state.config.REQUEST_CONN_TIMEOUT
    if request_write_timeout < 0:
        request_write_timeout = app.state.config.REQUEST_WRITE_TIMEOUT
    if response_read_timeout < 0:
        response_read_timeout = app.state.config.RESPONSE_READ_TIMEOUT
    if request_retry_count < 0:
        request_retry_count = app.state.config.REQUEST_RETRY_COUNT
    if request_retry_delay_initial_seconds < 0:
        request_retry_delay_initial_seconds = app.state.config.REQUEST_RETRY_DELAY_INITIAL_SECONDS

    timeout = httpx.Timeout(connect=request_conn_timeout, read=response_read_timeout,
                            write=request_write_timeout, pool=request_conn_pool_timeout)

    if "headers" in kwargs.keys():
        kwargs.get("headers")[CORRELATION_ID_HEADER_KEY_NAME] = correlation_id.get() or ""
    else:
        kwargs["headers"] = {
            CORRELATION_ID_HEADER_KEY_NAME: correlation_id.get() or ""
        }

    if current_user and "access_token" in current_user:
        if "headers" in kwargs.keys():
            kwargs.get("headers")["Authorization"] = f"Bearer {current_user.get('access_token')}"
        else:
            kwargs["headers"] = {
                "Authorization": f"Bearer {current_user.get('access_token')}"
            }

    if request_retry_count <= 0:
        response = await invoke_request(timeout, method, url, app, **kwargs)
        return response
    else:
        retries = 0
        current_delay = request_retry_delay_initial_seconds

        while retries <= request_retry_count:
            if retries > 0:
                app.state.logger.warn(f"【{app.title}】 \nURL: {method} {url} \nkwargs: {kwargs} \n重新嘗試送請求 (第 {retries} 次嘗試)...")

            try:
                # 使用 await 關鍵字等待異步請求完成
                response = await invoke_request(timeout, method, url, app, **kwargs)
                return response
            except GatewayTimeoutException as e:
                if retries < request_retry_count:
                    # 計算下一次的延遲時間：current_delay * 2^retries + 隨機抖動
                    sleep_time = current_delay * (app.state.config.REQUEST_RETRY_DELAY_FACTOR ** retries) + random.uniform(app.state.config.REQUEST_RETRY_DELAY_RANDOM_JITTER_MIN, app.state.config.REQUEST_RETRY_DELAY_RANDOM_JITTER_MAX)
                    app.state.logger.warn(f"【{app.title}】 \nURL: {method} {url} \nkwargs: {kwargs} \n等待 {sleep_time:.2f} 秒後重試...")
                    # 使用 asyncio.sleep 進行異步等待，不會阻塞主執行緒
                    await asyncio.sleep(sleep_time)
                    retries += 1
                else:
                    message = f"【{app.title}】 \nURL: {method} {url} \nkwargs: {kwargs} \n已達到最大重試次數 ({request_retry_count})，放棄發送請求。"
                    app.state.logger.warn(message)
                    raise  # 重新拋出最後一個異常


async def invoke_webhook_message_api(app: FastAPI, message: str):
    payload = {"text": message}
    response = await async_request(app, "POST", app.state.config.WEBHOOK_BASE_URL, request_retry_count=0,
                                   request_retry_delay_initial_seconds=0.0, json=payload)
    response.raise_for_status()
    return response


async def send_webhook_message(app: FastAPI, message: str):
    if not app.state.config.WEBHOOK_BASE_URL:
        app.state.logger.warn(f"Skip notify webhook url is null")
        return None

    retry_count = app.state.config.WEBHOOK_RETRY_COUNT
    if retry_count <= 0:
        try:
            response = await invoke_webhook_message_api(app, message)
            return response
        except Exception as e:
            app.state.logger.warn(f"Notify failure, Exception:{e}")
    else:
        retries = 0
        current_delay = app.state.config.WEBHOOK_RETRY_DELAY_INITIAL_SECONDS

        while retries <= retry_count:
            if retries > 0:
                app.state.logger.warn(f"重新嘗試發送訊息 (第 {retries} 次嘗試)...")

            try:
                # 使用 await 關鍵字等待異步請求完成
                response = await invoke_webhook_message_api(app, message)
                return response
            except httpx.HTTPStatusError as e:
                if e.response.status_code == 429:
                    app.state.logger.warn(f"收到 429 錯誤：{e.response.status_code} - {e.response.text}")
                    if retries < retry_count:
                        # 計算下一次的延遲時間：current_delay * 2^retries + 隨機抖動
                        sleep_time = current_delay * (app.state.config.WEBHOOK_RETRY_DELAY_FACTOR ** retries) + random.uniform(app.state.config.WEBHOOK_RETRY_DELAY_RANDOM_JITTER_MIN, app.state.config.WEBHOOK_RETRY_DELAY_RANDOM_JITTER_MAX)
                        app.state.logger.warn(f"等待 {sleep_time:.2f} 秒後重試...")
                        # 使用 asyncio.sleep 進行異步等待，不會阻塞主執行緒
                        await asyncio.sleep(sleep_time)
                        retries += 1
                    else:
                        app.state.logger.warn(
                            f"已達到最大重試次數 ({app.state.config.WEBHOOK_RETRY_COUNT})，放棄發送訊息。")
                        raise  # 重新拋出最後一個異常
                else:
                    app.state.logger.warn(f"發生其他 HTTP 錯誤：{e.response.status_code} - {e.response.text}")
                    raise  # 對於非 429 錯誤，直接拋出

            except httpx.RequestError as e:
                # 處理網絡錯誤，例如 DNS 查找失敗、連接超時等
                app.state.logger.warn(f"發生網絡錯誤：{e}")
                if retries < retry_count:
                    sleep_time = current_delay * (2 ** retries) + random.uniform(0, 0.5)
                    app.state.logger.warn(f"等待 {sleep_time:.2f} 秒後重試...")
                    await asyncio.sleep(sleep_time)
                    retries += 1
                else:
                    app.state.logger.warn(f"已達到最大重試次數 ({retry_count})，放棄發送訊息。")
                    raise

        return None  # 如果重試次數用盡仍未成功



# -*- coding: UTF-8 -*-


DATE_FMT = "%Y-%m-%d"
TIME_FMT = "%H:%M"
DT_FMT = f"{DATE_FMT} {TIME_FMT}:%S"

DEFAULT_LOGGER_NAME = "uvicorn"
LOG_DT_FMT = f"{DT_FMT},%03d"
LOG_FMT_NO_DT = "[%(levelname)7s][%(correlation_id)s][%(name)s][%(filename)s::%(funcName)s(%(lineno)s)]: %(message)s"
LOG_FMT = "[%(asctime)s]" + LOG_FMT_NO_DT

ARR_EXPORT_SHORT_FMT = "MM/DD HH:mm"
ARR_EXPORT_DATE_FMT = "YYYY/MM/DD"
ARR_EXPORT_DATE_ONLY_FMT = "MM/DD"
ARR_DATE_FMT = "YYYY-MM-DD"
ARR_DATE_FMT_NO_DASH = "YYYYMMDD"
ARR_EXPORT_DATETIME_FMT = "YYYY/MM/DD HH:mm"
ARR_TIME_FMT = "HH:mm:ss"
ARR_TIME_ONLY_FMT = "HH:mm"
ARR_RFC3339_DT_FMT = f"{ARR_DATE_FMT}T{ARR_TIME_FMT}ZZ"
ARR_DATETIME_FMT_NO_DASH = "YYYYMMDDHHmmss"
ARR_HUMAN_READ_FMT = f"{ARR_DATE_FMT} {ARR_TIME_FMT}"
ARR_IMPORT_TS_FMT = "YYYYMMDDHHmmss.SSSSSS"
ARR_STR_DAY = "day"

STR_EMPTY = ""
STR_SPACE = " "
STR_DASH = "-"
STR_UNDERSCORE = "_"
STR_SLASH = "/"
STR_COMMA = ","

SECONDS_PER_YEAR = 30758400
SECONDS_PER_DAY = 86400
SECONDS_PER_MIN = 60

DEF_PAGE_NO = 1
DEF_PAGE_SIZE = 15
DEF_ORDER_BY = 'id'

APS_EVENT_CODE = {
    1: 'EVENT_SCHEDULER_STARTED',
    2: 'EVENT_SCHEDULER_SHUTDOWN',
    4: 'EVENT_SCHEDULER_PAUSED',
    8: 'EVENT_SCHEDULER_RESUMED',
    16: 'EVENT_EXECUTOR_ADDED',
    32: 'EVENT_EXECUTOR_REMOVED',
    64: 'EVENT_JOBSTORE_ADDED',
    128: 'EVENT_JOBSTORE_REMOVED',
    256: 'EVENT_ALL_JOBS_REMOVED',
    512: 'EVENT_JOB_ADDED',
    1024: 'EVENT_JOB_REMOVED',
    2048: 'EVENT_JOB_MODIFIED',
    4096: 'EVENT_JOB_EXECUTED',
    8192: 'EVENT_JOB_ERROR',
    16384: 'EVENT_JOB_MISSED',
    32768: 'EVENT_JOB_SUBMITTED',
    65536: 'EVENT_JOB_MAX_INSTANCES'
}

CORRELATION_ID_HEADER_KEY_NAME="X-Request-ID"

STR_DELIMIT = '$^$'


from fastapi import status
from .base_exception import InternalBaseException


class BadGatewayException(InternalBaseException):
    code = "error_bad_gateway"
    message = "請求失敗"

    def __init__(self, message: str = None, **kwargs):
        _message = message or self.message
        super().__init__(status.HTTP_502_BAD_GATEWAY, self.code, _message, **kwargs)


class GatewayTimeoutException(InternalBaseException):
    code = "error_gateway_timeout"
    message = "請求逾時"

    def __init__(self, message: str = None, **kwargs):
        _message = message or self.message
        super().__init__(status.HTTP_504_GATEWAY_TIMEOUT, self.code, _message, **kwargs)


class DatabaseInitializeFailureException(InternalBaseException):
    code = "error_database_initialize"
    message = "資料庫初始化失敗"

    def __init__(self, message: str = None, **kwargs):
        _message = message or self.message
        super().__init__(status.HTTP_500_INTERNAL_SERVER_ERROR, self.code, _message, **kwargs)


class DatabaseConnectFailureException(InternalBaseException):
    code = "error_database_connect"
    message = "資料庫連線失敗"

    def __init__(self, message: str = None, **kwargs):
        _message = message or self.message
        super().__init__(status.HTTP_500_INTERNAL_SERVER_ERROR, self.code, _message, **kwargs)


class NoChangeException(InternalBaseException):
    code = "error_no_change"
    message = "資料未異動"

    def __init__(self, message: str = None, **kwargs):
        _message = message or self.message
        super().__init__(status.HTTP_200_OK, self.code, _message, **kwargs)


class CustomerValidationException(InternalBaseException):
    code = "error_validation"
    message = "無效參數或格式不符"

    def __init__(self, message: str = None, **kwargs):
        _message = message or self.message
        super().__init__(status.HTTP_422_UNPROCESSABLE_ENTITY, self.code, _message, **kwargs)



import arrow
from datetime import datetime
from typing import Optional, Dict

import dictdiffer
from pydantic import BaseModel


class Operate(BaseModel):
    add: Optional[dict] = None
    remove: Optional[dict] = None
    change: Optional[dict] = None

    @classmethod
    async def generate_operate(cls, original: dict = None, compare: dict = None):
        if original:
            original = await cls.remove_ignore_field(original)
            await cls.convert_datetime_timezone_utc_field(original)
        else:
            original = {}

        if compare:
            compare = await cls.remove_ignore_field(compare)
            await cls.convert_datetime_timezone_utc_field(compare)
        else:
            compare = {}

        diff_result = {'add': {}, 'remove': {}, 'change': {}}
        for diff in list(dictdiffer.diff(original, compare)):
            diff = (list(diff))
            if diff[0] == 'add' or diff[0] == 'remove':
                diff[2] = list(diff[2])
                for j in range(0, len(diff[2])):
                    diff[2][j] = list(diff[2][j])
            if diff[1] == 'change':
                diff[2] = list(diff[2])

            if diff[0] == 'change':
                temp_key = diff[1]
                if type(diff[1]) == list:
                    temp_key = "_".join([str(temp_x) for temp_x in list(diff[1])])
                diff_result['change'][temp_key] = {'orig': diff[2][0], 'new': diff[2][1]}

            if diff[0] == 'add' or diff[0] == 'remove':
                for j in range(0, len(diff[2])):
                    if str(diff[2][j][0]).isdigit():
                        diff_result[diff[0]][f"{diff[1]}_{diff[2][j][0]}"] = diff[2][j][1]
                    else:
                        diff_result[diff[0]][diff[2][j][0]] = diff[2][j][1]

        final_diff_result = cls.split_keys(diff_result)
        return Operate(**final_diff_result)

    @classmethod
    async def remove_ignore_field(cls, model_dict: dict):
        return {k: v for k, v in model_dict.items() if k not in ['create_time', 'update_time']}

    @classmethod
    async def convert_datetime_timezone_utc_field(cls, model_dict: dict):
        # 統一使用arrow.get取代datetime，使其包含tzinfo，避免diff因時區有無判斷錯誤
        for k, v in model_dict.items():
            if isinstance(v, datetime):
                model_dict[k] = arrow.get(v).datetime

    @classmethod
    def split_keys(cls, diff_result: Dict):
        result = {}
        for key, value in diff_result.items():
            key = str(key)
            if '.' in key:
                parts = key.split('.')
                current_level = result
                for i, part in enumerate(parts):
                    if i == len(parts) - 1:
                        current_level[part] = cls.split_keys(value) if isinstance(value, dict) else value
                    else:
                        if part not in current_level:
                            current_level[part] = {}
                        current_level = current_level[part]
            else:
                result[key] = cls.split_keys(value) if isinstance(value, dict) else value
        return result



import logging
import time

from collections import defaultdict
from fastapi import FastAPI, Request
from starlette.middleware.base import BaseHTTPMiddleware
from ..const import CORRELATION_ID_HEADER_KEY_NAME


class LogRequestMiddleware(BaseHTTPMiddleware):
    def __init__(self, app: FastAPI, logger: logging.Logger):
        super().__init__(app)
        self.app = app
        self.logger = logger

    async def dispatch(self, request: Request, call_next):
        # 记录请求的URL和参数
        url = request.url.path
        method = request.method
        headers = request.headers
        request_id = headers.get(CORRELATION_ID_HEADER_KEY_NAME, "")

        query_params = request.query_params
        temp = defaultdict(list)
        for key, value in query_params.multi_items():
            temp[key].append(value)
        # 如果只有一個值就轉成單一值
        params = {k: v[0] if len(v) == 1 else v for k, v in temp.items()}

        body = await request.body()
        self.logger.info(f"[Request id: {request_id}] \nURL: {method} {url} \nParams: {params} \nBody: {body} \nHeaders: {headers} \nstart processing...")

        # 记录请求处理时间
        start_time = time.time()
        response = await call_next(request)
        process_time = time.time() - start_time

        self.logger.info(f"[Request id: {request_id}] \nURL: {method} {url} \nParams: {params} \nBody: {body} \nCompleted in {process_time:.4f} seconds")
        return response



# -*- coding: utf-8 -*-
import json

import arrow

from io import BytesIO
from PIL import Image
from PIL import UnidentifiedImageError
import pillow_heif

from .base_config import BaseConfig
from .const import STR_EMPTY, ARR_EXPORT_DATETIME_FMT, STR_DASH


def is_today(time, system_time_zone):
    if not time:
        return False

    return arrow.now(system_time_zone).floor("day") == arrow.get(time).to(system_time_zone).floor("day")


def get_tz_day_boundary(date_time, time_zone, out_tz="UTC"):
    """
    傳入date_time，取其在dt_tz時區的當天的floor與ceil，以out_tz時區回傳
    比對區間需使用gte/lte
    """
    date_time = arrow.get(date_time) if date_time else arrow.get()
    tz_time = date_time.to(time_zone)
    return tz_time.floor("day").to(out_tz), tz_time.ceil("day").to(out_tz)


def get_tz_month_boundary(date_time, time_zone, out_tz="UTC"):
    """
    傳入date_time，取其在dt_tz時區的當月的floor與ceil，以out_tz時區回傳
    比對區間需使用gte/lte
    """
    date_time = arrow.get(date_time) if date_time else arrow.get()
    tz_time = date_time.to(time_zone)
    return tz_time.floor("month").to(out_tz), tz_time.ceil("month").to(out_tz)


def timestamp_interval(start, end, interval_sec):
    while start < end:
        yield start
        start += interval_sec


def export_time_format(date, time_zone, fmt=ARR_EXPORT_DATETIME_FMT):
    if not date:
        return STR_EMPTY

    return arrow.get(date).to(time_zone).format(fmt)


def update_dict_with_cast(curr_settings: BaseConfig, new_conf: dict):
    if issubclass(type(curr_settings), BaseConfig):
        for key, value in new_conf.items():
            if hasattr(curr_settings, key):
                key_type = type(getattr(curr_settings, key))
                cast_func = key_type if key_type in (str, int) else json.loads
                setattr(curr_settings, key, cast_func(value))


def sanitize_plate_no(plate_no):
    return plate_no.replace(STR_DASH, STR_EMPTY)


async def open_image_stream(stream):
    try:
        return Image.open(stream)
    except UnidentifiedImageError as err:
        await stream.seek(0)
        bytes_data = await stream.read()
        return read_heif_to_pil(bytes_data)


def read_heif_to_pil(heif_file):
    heif_image = pillow_heif.read_heif(heif_file)
    return Image.frombytes(
        heif_image.mode,
        heif_image.size,
        heif_image.data,
        "raw",
        heif_image.mode,
        heif_image.stride,
    )


def convert_pil_to_bytes(pil_image, format='JPEG'):
    target_bytes = BytesIO()
    if pil_image.mode in ('RGBA', 'P'):
        pil_image = pil_image.convert('RGB')
    pil_image.save(target_bytes, format=format, quality=95)
    return target_bytes
